#!/usr/bin/env python

#    DiffColor.py colors and line-wraps the output of Python's difflib
#    module.
#
#    Copyright (C) 2010  Winston C. Yang (winston@cs.wisc.edu)
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see
#    <http://www.gnu.org/licenses/>.

"""Color the output of Python's difflib module using ANSI escape
sequences.

The output is suitable for terminal (command line) windows in Unix or
Mac OS X.

Type `python3.1 DiffColor.py --help' for more information."""

# Created by Winston C. Yang
# Created on or before 2007-10-28
# winston@cs.wisc.edu

import difflib
from optparse import OptionParser
import os
import re
import sys
import time

# A class for Ansi style and color codes.
class Ansi:

    s = "\033[{0}m"

    style = {
        "default"      : s.format(0),
        "underline"    : s.format(4),
        "reverse"      : s.format(7),
        "no_underline" : s.format(24)
        }

    # Foreground colors.
    fg = {
        "black"   : s.format(30),
        "red"     : s.format(31),
        "green"   : s.format(32),
        "yellow"  : s.format(33),
        "blue"    : s.format(34),
        "magenta" : s.format(35),
        "cyan"    : s.format(36),
        "white"   : s.format(37)
         }

    # Background colors.
    bg = {
        "black"   : s.format(40),
        "red"     : s.format(41),
        "green"   : s.format(42),
        "yellow"  : s.format(43),
        "blue"    : s.format(44),
        "magenta" : s.format(45),
        "cyan"    : s.format(46),
        "white"   : s.format(47)
        }

# A class for holding data about each "local diff" ("line-by-line
# change") between the files.
class LocalDiff:

    def __init__(self, pattern, n1, L1, s1, n2, L2, s2):

        self.pattern = pattern
        self.line_no1 = n1
        self.line1 = L1
        self.symbols1 = s1
        self.line_no2 = n2
        self.line2 = L2
        self.symbols2 = s2

# A class for holding data about the colored output (with Ansi escape
# sequences) of each "local diff."
class OutputLine:

    def __init__(self, n1, t1, n2, t2):

        self.line_no1 = n1
        self.text1 = t1
        self.line_no2 = n2
        self.text2 = t2

class DiffColor:

    """Color the output of Python's difflib module using ANSI escape
    sequences."""

    def __init__(self):

        # Set foreground and background colors.
        self.fg0 = Ansi.fg["black"]
        self.bg0 = Ansi.bg["white"]
        self.bg1 = Ansi.bg["cyan"] # Ansi.bg["red"]
        self.bg2 = Ansi.bg["magenta"] # Ansi.bg["green"]
        self.bgx = Ansi.bg["yellow"]

        # difflib has "symbol lines" in its diff output, to indicate
        # differences in the line preceding the "symbol line." We will
        # use each "symbol line" to color the characters in the line
        # preceding the "symbol line."
        self.format = {
            " ": self.bg0,
            "\t": self.bg0,
            "\n": self.bg0,
            "-": self.bg1,
            "+": self.bg2,
            "^": self.bgx,
            "!": self.bgx
            }

        # The output will be of the form:
        # Diff set 1
        # File 1 line number, margin, file 1 text, ...,
        # file 2 line number, margin, file 2 text
        # :
        # :
        self.margin = " "

        # Get command-line arguments.
        usage = "Usage: [python3.1] DiffColor.py [options] file1 file2"
        parser = OptionParser(usage=usage)
        parser.add_option("--columns",
                          dest="columns",
                          type="int",
                          default=80,
                          help="width of output")
        parser.add_option("--tab",
                          dest="tab",
                          type="int",
                          default=8,
                          help="replace each tab by this number of spaces")
        (options, args) = parser.parse_args()

        # Check for exactly two file names.
        if len(args) != 2:
            print("Enter two file names")
            sys.exit(1)

        self.file_name1 = args[0]
        self.file_name2 = args[1]

        def check_file_exists(file):
            try:
                f = open(file)
            except IOError:
                print("File does not exist: " + file)
                sys.exit(1)
            else:
                f.close()

        check_file_exists(self.file_name1)
        check_file_exists(self.file_name2)

        t = 64
        if options.columns < t:
            print("--columns value must be an integer greater than "\
                      "or equal to {}, to avoid small windows"\
                      .format(t))
            sys.exit(1)

        t = 1
        if options.tab < t:
            print("--tab value must be an integer greater than "\
                      "or equal to {}".format(t))
            sys.exit(1)

        self.options = options

    def run(self):

        """Diff the files specified on the command line."""

        def get_lines(file_name):
            f = open(file_name)
            try:
                t = f.read()
            except UnicodeDecodeError as e:
                print("Unicode decode error in file {}:\n{}"
                      .format(file_name, e))
                sys.exit(1)
            # Replace tabs by possibly more tabs.
            t = re.sub("\t", self.options.tab * "\t", t)
            return t.splitlines()

        lines1 = get_lines(self.file_name1)
        lines2 = get_lines(self.file_name2)
        lines = list(difflib.Differ().compare(lines1, lines2))

        # Determine if the files are different.
        if len(lines) == len(lines1) == len(lines2):
            return

        # File modification time header.
        def get_modification_time(path):
            t = time.localtime(os.path.getmtime(path))
            return time.strftime("%a %Y-%m-%d %H:%M:%S", t)

        def print_file_header(symbol, file_name, file_name_width, color):
            t = "{0} {1} {2}".format\
                (3*symbol, file_name.ljust(file_name_width),
                 get_modification_time(file_name))
            t = t.ljust(self.options.columns)
            print(self.fg0 + color + t + Ansi.style["default"])

        m = max(len(self.file_name1), len(self.file_name2))
        print_file_header("-", self.file_name1, m, self.bg1)
        print_file_header("+", self.file_name2, m, self.bg2)

        # Reverse the foreground and background colors for the header.
        def print_diff_set_header(text):
            t = Ansi.style["reverse"] + self.fg0 + self.bg0\
                + "{:<" + str(self.options.columns) + "}"\
                + Ansi.style["default"]
            print(t.format(text))

        # Pane width.
        c1 = len(str(len(lines1))) # File 1 line number width.
        c2 = len(str(len(lines2))) # File 2 line number width.
        m = len(self.margin) # Margin width (between line number and code).
        pane_width = (self.options.columns - (c1 + c2 + m + m))//2

        def print_diff_set_line(no1, line1, no2, line2):
            s = self.bg0 + self.margin
            t = self.fg0\
                + self.bg1 + "{:" + str(c1) + "}" + s + "{}"\
                + self.bg2 + "{:" + str(c2) + "}" + s + "{}"\
                + Ansi.style["default"]
            print(t.format(no1, line1, no2, line2))

        def make_tabs_visible(text):
            u1 = Ansi.style["underline"]
            u2 = Ansi.style["no_underline"]
            s = re.sub("(\t+)", u1 + "\\1" + u2, text)
            return s.replace("\t", " ")

        local_diffs = self.__get_local_diffs(lines, pane_width)

        i = 1
        # Group consecutive local diffs that are changes into a diff
        # set. This may improve comprehensibility.
        continue_diff_set = False
        for local_diff in local_diffs:

            # The lines in the files are the same. End any ongoing
            # diff set.
            if local_diff.pattern == " ":
                continue_diff_set = False
                continue

            # If there is no ongoing diff set, start one.
            if not continue_diff_set:
                continue_diff_set = True
                print_diff_set_header("Diff set " + str(i))
                i += 1

            # Print lines of colored output about a local diff.
            for o in self.__get_output_lines(local_diff, pane_width):
                n1 = o.line_no1
                t1 = make_tabs_visible(o.text1)
                n2 = o.line_no2
                t2 = make_tabs_visible(o.text2)
                print_diff_set_line(n1, t1, n2, t2)

    def __get_local_diffs(self, lines, pane_width):

        # Line numbers.
        i = 0
        i1 = 1
        i2 = 1

        L = lines
        symbols = "".join([line[0] for line in L])
        assert set(symbols).issubset(set(" -+?"))

        local_diffs = []
        patterns = (" ", "-?+?", "-?+", "-+?", "-", "+")
        length = len(L)

        while i < length:

            for j, p in enumerate(patterns):
                if symbols[i:i + len(p)] == p:
                    if j == 0:
                        args = (p, i1, L[i], "", i2, L[i], "")
                    elif j == 1:
                        args = (p, i1, L[i], L[i + 1], i2, L[i + 2], L[i + 3])
                    elif j == 2:
                        args = (p, i1, L[i], L[i + 1], i2, L[i + 2], "")
                    elif j == 3:
                        args = (p, i1, L[i], "", i2, L[i + 1], L[i + 2])
                    elif j == 4:
                        args = (p, i1, L[i], "", "", "", "")
                    elif j == 5:
                        args = (p, "", "", "", i2, L[i], "")
                    else:
                        # Should not get here.
                        raise Exception
                    break

            local_diffs.append(LocalDiff(*args))
            i1 += (" " in p) + ("-" in p)
            i2 += (" " in p) + ("+" in p)
            i += len(p)

        return local_diffs

    def __get_output_lines(self, local_diff, width):

        # Get a substring at a given position and with a given width,
        # left-justified.
        def slice(string, i):
            return string[i:i + width].ljust(width)

        # Get a range for iterating over strings of a given width. Put
        # "width" in the max so that there will be at least one
        # iteration.
        def range_w(string1, string2=""):
            return range(0, max(width, len(string1), len(string2)), width)

        # Drop at most the first two characters (note: there might not
        # be even two characters---the lines and symbol lines could be
        # empty). The first two characters may be two spaces, a plus
        # sign and a space, or a minus sign and a space.
        line1, symbols1 = local_diff.line1[2:], local_diff.symbols1[2:]
        line2, symbols2 = local_diff.line2[2:], local_diff.symbols2[2:]

        # Filler space for one file's output, if the file has no line
        # corresponding to a line in the other file.
        spaces = width * " "

        output_lines = []

        # A line is in file 1 but not in file 2.
        if local_diff.line_no1 and not local_diff.line_no2:
            for i in range_w(line1):
                t = slice(line1, i)
                # For now, the line numbers are empty. We'll add them later.
                output_lines.append(OutputLine("", self.bg1 + t, "", spaces))

        # A line is not in file 1 but in file 2.
        elif not local_diff.line_no1 and local_diff.line_no2:
            for i in range_w(line2):
                t = slice(line2, i)
                # For now, the line numbers are empty. We'll add them later.
                output_lines.append(OutputLine("", spaces, "", self.bg2 + t))

        # The lines in file 1 and file 2 conflict.
        else:

            def get_colored_text(line, symbols):
                # Insert the formats for the symbols into the
                # corresponding indices in the line. Traverse the
                # indices in reverse. Else, after the first insertion,
                # any other indices will be out of synch.
                indices = [0] + [s for s in range(1, len(symbols))\
                                     if symbols[s] != symbols[s - 1]]
                indices.reverse()

                chars = list(line)
                for i in indices:
                    chars[i:i] += self.format[symbols[i]]
                return "".join(chars)

            for i in range_w(line1, line2):
                f1 = get_colored_text(slice(line1, i), slice(symbols1, i))
                f2 = get_colored_text(slice(line2, i), slice(symbols2, i))
                # For now, the line numbers are empty. We'll add them later.
                output_lines.append(OutputLine("", f1, "", f2))

        assert output_lines

        # Make the first output line have line numbers.
        t = output_lines[0]
        t.line_no1 = local_diff.line_no1
        t.line_no2 = local_diff.line_no2

        return output_lines

if __name__ == "__main__":
    DiffColor().run()
